@@ -10,12 +10,15 @@
 -}
 
 module Pugs.Compile where
+import Pugs.AST
+import Pugs.Internals
 import Pugs.Compile.Pugs (genPugs)
-import Pugs.Compile.Parrot (genPIR)
+import Pugs.Compile.Parrot (genIMC)
 import Pugs.Compile.Haskell (genGHC)
 
-compile "Haskell" = genGHC
-compile "Pugs" = genPugs
-compile "Parrot" = genPIR
-compile s = const (error $ "Cannot compile to " ++ s)
+compile :: String -> Env -> IO String
+compile "Haskell" env = fmap vCast $ runEvalIO env genGHC
+compile "Pugs"    env = fmap vCast $ runEvalIO env genPugs
+compile "Parrot"  env = fmap vCast $ runEvalIO env genIMC
+compile s _ = fail $ "Cannot compile to " ++ s
 
