@@ -3,9 +3,11 @@ use strict;
 use warnings;
 use Cwd;
 
+my $version_h = shift || Cwd::cwd() . "src/Pugs/pugs_version.h";
 my $base = shift || Cwd::cwd();
-my $version_h = "$base/src/pugs_version.h";
-my $svn_entries = ".svn/entries";
+my $svn_entries = "$base/.svn/entries";
+
+print "Writing version from $svn_entries to $version_h\n";
 
 open OUT, "> $version_h" or die $!;
 print OUT "#undef PUGS_SVN_REVISION\n";
@@ -15,7 +17,8 @@ if (-r $svn_entries) {
         /^ *committed-rev=.(\d+)./ or next;
         print OUT "#define PUGS_SVN_REVISION $1\n";
         # rebuild Help.hs to show new revision number
-        unlink "$base/src/Help.hi"; 
+        unlink "$base/src/Pugs/Help.hi";
+        unlink "$base/src/Pugs/Help.o";
         exit;
     }
 }
